/***************************************************************************

  Mini-Pac Copyright (C) 1999 by Michael Balfour.  All Rights Reserved.


  Pac-Man Copyright (C) 1980 by Namco, Ltd.  All Rights Reserved.


  Multi-Z80 32 Bit emulator 
  Copyright (C) 1996-1999 Neil Bradley, All rights reserved
  Author      : Neil Bradley (neil@synthcom.com)
  Distribution: ftp://ftp.synthcom.com/pub/emulators/cpu/makez80.zip (latest)

 ***************************************************************************/

// INCLUDES ///////////////////////////////////////////////////////////////////

#include "XPac.h"

char *TileGraphicsRomName;
char *TileGraphicsRomName2 = NULL;
char *SpriteGraphicsRomName;
char *SpriteGraphicsRomName2 = NULL;


char *ColorPaletteRomName;
char *ColorLookupTableRomName;

char *SoundRomName;

DipSwitch *DipLook;
PortData *PortLook;

int CurrentSwitch = -1;

/*-----------------------------------------------------------------------------
   Name: Interpreter()
   Desc: Construct our main object
  -----------------------------------------------------------------------------*/

Interpreter::Interpreter()
{
	ofstream foo("xpac.dbg");
	int i;
	for (i=0; i<10; i++)
		foo << DipDefault[i] << "\n";
	foo.close();
	if (GameSelected == 0)
		throw ("Uhoh!");
	emuMemory = NULL;
	dirtyVideoMemory = NULL;

	tilePtr = NULL;
	spritePtr = NULL;
	spriteHFlipPtr = NULL;
	spriteVFlipPtr = NULL;
	spriteHVFlipPtr = NULL;

	origPalette = NULL;
	origColorLookup = NULL;

	// Init Emulation Memory
	switch (GameSelected)
	{
		case NAMCOPAC:
			TileGraphicsRomName = "namcopac.5e";
			SpriteGraphicsRomName = "namcopac.5f";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(NamcoPacRomList);
			DipLook = NamcoPacSwitches;
			PortLook = NamcoPacPorts;
			break;
		case MIDWAYPAC:
			TileGraphicsRomName = "pacman.5e";
			SpriteGraphicsRomName = "pacman.5f";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(MidwayPacRomList);
			break;
		case MSPACMAN:
			TileGraphicsRomName = "5e";
			SpriteGraphicsRomName = "5f";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(MsPacRomList);
			break;
		case PACPLUS:
			TileGraphicsRomName = "pacplus.5e";
			SpriteGraphicsRomName = "pacplus.5f";
			ColorPaletteRomName = "pacplus.7f";
			ColorLookupTableRomName = "pacplus.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(PacPlusRomList);
			break;
		case PACMOD:
			TileGraphicsRomName = "pacmanh.5e";
			SpriteGraphicsRomName = "pacmanh.5f";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(PacModRomList);
			break;
		case PACMANJP:
			TileGraphicsRomName = "chg1";
			TileGraphicsRomName2 = "chg2";
			SpriteGraphicsRomName = "chg3";
			SpriteGraphicsRomName2 = "chg4";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(PacManJpRomList);
			break;
		case HANGLY:
			TileGraphicsRomName = "hangly.5e";
			SpriteGraphicsRomName = "hangly.5f";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(HanglyRomList);
			break;
		case PUCKMAN:
			TileGraphicsRomName = "puckman.5e";
			SpriteGraphicsRomName = "puckman.5f";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(PuckManRomList);
			break;
		case MSPACATK:
			TileGraphicsRomName = "5e";
			SpriteGraphicsRomName = "5f";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(MsPacAtkRomList);
			break;
		case THEGLOB:
			TileGraphicsRomName = "glob.5e";
			SpriteGraphicsRomName = "glob.5f";
			ColorPaletteRomName = "glob.7f";
			ColorLookupTableRomName = "glob.4a";
			SoundRomName = "pacman.spr";
			InitEmuMemory(TheGlobRomList);
			break;
		case PIRANHA:
			TileGraphicsRomName = "pr5.cpu";
			TileGraphicsRomName2 = "pr7.cpu";
			SpriteGraphicsRomName = "pr6.cpu";
			SpriteGraphicsRomName2 = "pr8.cpu";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(PiranhaRomList);
			break;
		case CRUSH:
			TileGraphicsRomName = "tpa";
			TileGraphicsRomName2 = "tpc";
			SpriteGraphicsRomName = "tpb";
			SpriteGraphicsRomName2 = "tpd";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(CrushRomList);
			break;
		case PONPOKO:
			TileGraphicsRomName = "ppoko9.bin";
			SpriteGraphicsRomName = "ppoko10.bin";
			ColorPaletteRomName = "82s123.7f";
			ColorLookupTableRomName = "82s126.4a";
			SoundRomName = "82s126.1m";
			InitEmuMemory(PonpokoRomList);
			break;
		}
	
	// Select DIP Switches

	switch (GameSelected)
	{
		case MSPACMAN:
		case MSPACATK:
			DipLook = MsPacSwitches;
		case PACPLUS:
		case PACMOD:
		case PACMANJP:
		case HANGLY:
		case PUCKMAN:
		case MIDWAYPAC:
		case THEGLOB:
		case PIRANHA:
		case NAMCOPAC:
			DipLook = NamcoPacSwitches;
			PortLook = NamcoPacPorts;
			break;
		case CRUSH:
			DipLook = CrushSwitches;
			PortLook = NamcoPacPorts;
			break;
			
		//case PONPOKO:
		}


	// Init Sound
	InitSound();

	// Init Graphics
	InitGraphics();

	// Init Colors
	InitColors();

	// Init Sprite List
	InitSprites();

	// Init inputs and DIP switches
	InitInputPorts();
}

/*-----------------------------------------------------------------------------
   Name: ~Interpreter()
   Desc: Destroy our main object
  -----------------------------------------------------------------------------*/

Interpreter::~Interpreter()
{
	if (emuMemory != NULL)			free (emuMemory);
	if (dirtyVideoMemory != NULL)	free (dirtyVideoMemory);

	if (tilePtr != NULL)			free (tilePtr);
	if (spritePtr != NULL)			free (spritePtr);
	if (spriteHFlipPtr != NULL)		free (spriteHFlipPtr);
	if (spriteVFlipPtr != NULL)		free (spriteVFlipPtr);
	if (spriteHVFlipPtr != NULL)	free (spriteHVFlipPtr);

	if (origPalette != NULL)		free (origPalette);
	if (origColorLookup != NULL)	free (origColorLookup);
}

/*-----------------------------------------------------------------------------
   Name: CreateEmuMemory()
   Desc: Load code ROMs and create a memory space for the emulation
  -----------------------------------------------------------------------------*/
bool Interpreter::InitEmuMemory(CODE_ROM *romList)
{
	FILE *f;
	CODE_ROM *curRom;

	curRom = romList;

	/* Create the emulated game area */
	emuMemory = (UCHAR *)calloc(0x10000,1);

	/* Load the game ROMs */
	while (curRom->name != NULL)
	{
		f = fopen(curRom->name,"rb");
		if (f == NULL)
		{
			throw "Unable to locate game ROM";
		}

		if (fread(&(emuMemory[curRom->location]),curRom->size,1,f) != 1)
		{
			throw "Unable to read game ROM";
		}
		fclose(f);
        curRom++;
	}

	if (GameSelected == PACPLUS)
	{
		int i;

		for (i = 0; i < 0x4000; i++)
		{
			emuMemory[i] = PacPlusDecrypt(i, emuMemory[i]);
		}
	}


	dirtyVideoMemory = (UCHAR *)malloc(0x400);
	if (dirtyVideoMemory == NULL)
	{
		throw "Error allocating dirtyMemory block";
	}

	memset(dirtyVideoMemory,1,0x400);

	return true;
}

/*-----------------------------------------------------------------------------
   Name: InitInputPorts()
   Desc: Initialize input ports and DIP switches
  -----------------------------------------------------------------------------*/
void Interpreter::InitInputPorts(void)
{
	UCHAR data = 0x00;
	int CurrentNum = 0;
	int CurrentPort = -1;
	ofstream foo("xpac.db2");
	
	while (DipLook->mask != -2)
		{
			switch (DipLook->mask)
			{
				case -2:
					emuMemory[PortLook->location] = data;
					break;
				case -1:
					if (CurrentPort != -1)
					{
						emuMemory[PortLook->location] = data;
						PortLook++;
					}
					CurrentPort++;
					CurrentNum = -1;
					data = 0x00;
					data = PortLook->defaultmask;
					break;
				case 1: // DIPNAME
					CurrentSwitch++;
					
					foo << "NAME" << DipLook->name << "\n";
					if (DipDefault[CurrentSwitch] == -1)
					{
						DipDefault[CurrentSwitch] = (INT) DipLook->defaultvalue;
						foo << "DEFAULT\n";
						foo << (INT) DipLook->defaultvalue << "\n";
					}
					CurrentNum = -1;
					break;
				case 0: // DIPSETTING
					CurrentNum++;
					if (DipDefault[CurrentSwitch] == (CurrentNum + 1))
					{
						data |= DipLook->defaultvalue;
						foo << "FOUND SETTING - " << DipLook->defaultvalue << "\n";
					}
					break;
			}
			DipLook++;
		}
		emuMemory[PortLook->location] = data;
}

/*-----------------------------------------------------------------------------
   Name: InitSound()
   Desc: Load the sound ROM
  -----------------------------------------------------------------------------*/
void Interpreter::InitSound(void)
{
	FILE *f;
	int i;

	for (i = 0; i < NUM_SOUND_VOICES; i++ )
	{
		frequency[i] = 0;
		volume[i] = 0;
		wave[i] = 0;
	}
	
	f = fopen(SoundRomName,"rb");
	if (f == NULL)
	{
		throw "Error opening sound ROM";
	}

	if (fread(waveBuffer,0x100,1,f) != 1)
	{
		throw "Error reading sound ROM";
	}

	fclose(f);

	// Scale from a 0-F range to a 00-FF range.
	for (i = 0; i < 0x100; i++ )
	{
		waveBuffer[i] |= ((waveBuffer[i] & 0x0F) << 4);
	}
}

/*-----------------------------------------------------------------------------
   Name: InitGraphics()
   Desc: Initialize and decode graphics
  -----------------------------------------------------------------------------*/
void Interpreter::InitGraphics(void)
{
	FILE *f;
	UCHAR buffer[0x1000];

	/* Load the tiles */

	if (TileGraphicsRomName2 == NULL)
	{
		f = fopen(TileGraphicsRomName,"rb");
		if (f == NULL)
		{
			throw "Error opening tile graphics ROM";
		}

		if (fread(buffer,0x1000,1,f) != 1)
		{
			throw "Error reading tile graphics ROM";
		}

		fclose(f);
	}
	else
	{
		f = fopen(TileGraphicsRomName,"rb");
		if (f == NULL)
		{
			throw "Error opening tile graphics ROM";
		}

		if (fread(buffer,0x0800,1,f) != 1)
		{
			throw "Error reading tile graphics ROM";
		}
		fclose(f);

		f = fopen(TileGraphicsRomName2,"rb");
		if (f == NULL)
		{
			throw "Error opening tile graphics ROM";
		}

		if (fread(&(buffer[0x0800]),0x0800,1,f) != 1)
		{
			throw "Error reading tile graphics ROM";
		}

		fclose(f);
	}

	tilePtr = (UCHAR *)malloc(8*8*256);

	if (tilePtr == NULL)
	{
		throw "Error creating tile graphics memory";
	}

	DecodeTiles(buffer, tilePtr);

	/* Load the sprites */

	if (SpriteGraphicsRomName2 == NULL)
	{

		f = fopen(SpriteGraphicsRomName,"rb");
		if (f == NULL)
		{
			throw "Error opening sprite graphics ROM";
		}

		if (fread(buffer,0x1000,1,f) != 1)
		{
			throw "Error reading sprite graphics ROM";
		}

		fclose(f);
	}
	else
	{
		f = fopen(SpriteGraphicsRomName,"rb");
		if (f == NULL)
		{
			throw "Error opening sprite graphics ROM";
		}

		if (fread(buffer,0x0800,1,f) != 1)
		{
			throw "Error reading sprite graphics ROM";
		}

		fclose(f);

		f = fopen(SpriteGraphicsRomName2,"rb");
		if (f == NULL)
		{
			throw "Error opening sprite graphics ROM";
		}

		if (fread(&(buffer[0x0800]),0x0800,1,f) != 1)
		{
			throw "Error reading sprite graphics ROM";
		}

		fclose(f);
	}

	spritePtr        = (UCHAR *)malloc(16*16*64);
	spriteHFlipPtr   = (UCHAR *)malloc(16*16*64);
	spriteVFlipPtr   = (UCHAR *)malloc(16*16*64);
	spriteHVFlipPtr  = (UCHAR *)malloc(16*16*64);

	if ((spritePtr == NULL)      || (spriteHFlipPtr == NULL) || 
		(spriteVFlipPtr == NULL) || (spriteHVFlipPtr == NULL))
	{
		throw "Error creating sprite graphics memory";
	}

	DecodeSprites(buffer, spritePtr);
	FlipSpritesHoriz(spritePtr, spriteHFlipPtr);
	FlipSpritesVert(spritePtr, spriteVFlipPtr);
	FlipSpritesHorizVert(spritePtr, spriteHVFlipPtr);

}

/*-----------------------------------------------------------------------------
   Name: InitColors()
   Desc: Initialize and load colors
  -----------------------------------------------------------------------------*/
void Interpreter::InitColors(void)
{
	FILE *f;

    origPalette = (UCHAR *)malloc(0x20);
	origColorLookup = (UCHAR *)malloc(0x100);

	if ((origPalette == NULL) || (origColorLookup == NULL))
	{
		throw "Error creating color memory";
	}

	/* Load the color palette and lookup table */
	f = fopen(ColorPaletteRomName,"rb");
	if (f == NULL)
	{
		throw "error opening palette ROM";
	}
	if (fread(origPalette,0x20,1,f) != 1)
	{
		throw "error reading palette ROM";
	}
	fclose(f);

	f = fopen(ColorLookupTableRomName,"rb");
	
	if (f == NULL)
	{
		throw "error opening color lookup ROM";
	}
	if (fread(origColorLookup,0x100,1,f) != 1)
	{
		throw "error reading color lookup ROM";
	}
	fclose(f);
}

/*-----------------------------------------------------------------------------
   Name: DecodeTiles()
   Desc: Decode graphics from original buffer
  -----------------------------------------------------------------------------*/
void Interpreter::DecodeTiles(UCHAR *srcBuffer, UCHAR *destBuffer)
{
	UCHAR *s;
	UCHAR *d;

	for (int tile = 0; tile < 256; tile++)
	{
		s = &(srcBuffer[tile * 16]);
		d = &(destBuffer[tile * (8 * 8)]);

		for (int y = 7; y >= 0; y--)
		{
			d[8*0 + y] = ((s[8] & 0x80) >> 6) | ((s[8] & 0x08) >> 3);
			d[8*1 + y] = ((s[8] & 0x40) >> 5) | ((s[8] & 0x04) >> 2);
			d[8*2 + y] = ((s[8] & 0x20) >> 4) | ((s[8] & 0x02) >> 1);
			d[8*3 + y] = ((s[8] & 0x10) >> 3) | ((s[8] & 0x01) >> 0);

			d[8*4 + y] = ((s[0] & 0x80) >> 6) | ((s[0] & 0x08) >> 3);
			d[8*5 + y] = ((s[0] & 0x40) >> 5) | ((s[0] & 0x04) >> 2);
			d[8*6 + y] = ((s[0] & 0x20) >> 4) | ((s[0] & 0x02) >> 1);
			d[8*7 + y] = ((s[0] & 0x10) >> 3) | ((s[0] & 0x01) >> 0);

			s++;
		}
	}
}

/*-----------------------------------------------------------------------------
   Name: DecodeSprites()
   Desc: Decode graphics from original buffer
  -----------------------------------------------------------------------------*/
void Interpreter::DecodeSprites(UCHAR *srcBuffer, UCHAR *destBuffer)
{
	UCHAR *s;
	UCHAR *d;

	for (int sprite = 0; sprite < 64; sprite++)
	{
		s = &(srcBuffer[sprite * 64]);
		d = &(destBuffer[sprite * (16 * 16)]);

		for (int y = 15; y >= 0; y--)
		{
			if (s == &(srcBuffer[(sprite * 64) + 0x08]))
				s = &(srcBuffer[(sprite * 64) + 0x20]);

			d[16*0  + y] = ((s[8] & 0x80) >> 6) | ((s[8] & 0x08) >> 3);
			d[16*1  + y] = ((s[8] & 0x40) >> 5) | ((s[8] & 0x04) >> 2);
			d[16*2  + y] = ((s[8] & 0x20) >> 4) | ((s[8] & 0x02) >> 1);
			d[16*3  + y] = ((s[8] & 0x10) >> 3) | ((s[8] & 0x01) >> 0);

			d[16*4  + y] = ((s[16] & 0x80) >> 6) | ((s[16] & 0x08) >> 3);
			d[16*5  + y] = ((s[16] & 0x40) >> 5) | ((s[16] & 0x04) >> 2);
			d[16*6  + y] = ((s[16] & 0x20) >> 4) | ((s[16] & 0x02) >> 1);
			d[16*7  + y] = ((s[16] & 0x10) >> 3) | ((s[16] & 0x01) >> 0);
	
			d[16*8  + y] = ((s[24] & 0x80) >> 6) | ((s[24] & 0x08) >> 3);
			d[16*9  + y] = ((s[24] & 0x40) >> 5) | ((s[24] & 0x04) >> 2);
			d[16*10 + y] = ((s[24] & 0x20) >> 4) | ((s[24] & 0x02) >> 1);
			d[16*11 + y] = ((s[24] & 0x10) >> 3) | ((s[24] & 0x01) >> 0);
	
			d[16*12 + y] = ((s[0] & 0x80) >> 6) | ((s[0] & 0x08) >> 3);
			d[16*13 + y] = ((s[0] & 0x40) >> 5) | ((s[0] & 0x04) >> 2);
			d[16*14 + y] = ((s[0] & 0x20) >> 4) | ((s[0] & 0x02) >> 1);
			d[16*15 + y] = ((s[0] & 0x10) >> 3) | ((s[0] & 0x01) >> 0);

			s++;
		}
	}
}

/*-----------------------------------------------------------------------------
   Name: FlipSpritesHoriz()
   Desc: Take decoded sprites and horizontally flip them
  -----------------------------------------------------------------------------*/
void Interpreter::FlipSpritesHoriz(UCHAR *srcBuffer, UCHAR *flipBuffer)
{
	UCHAR *s;
	UCHAR *f;

	for (int sprite = 0; sprite < 64; sprite++)
	{
		int spriteOffset;

		spriteOffset = sprite * (16 * 16);

		for (int y = 15; y >= 0; y--)
		{
			s = &(srcBuffer[spriteOffset + (16 * y)]);
			f = &(flipBuffer[spriteOffset + (16 * y)]);

			for (int x = 15; x >= 0; x--)
			{
				f[x] = s[15 - x];				
			}
		}
	}
}

/*-----------------------------------------------------------------------------
   Name: FlipSpritesVert()
   Desc: Take decoded sprites and vertically flip them
  -----------------------------------------------------------------------------*/
void Interpreter::FlipSpritesVert(UCHAR *srcBuffer, UCHAR *flipBuffer)
{
	UCHAR *s;
	UCHAR *f;

	for (int sprite = 0; sprite < 64; sprite++)
	{
		int spriteOffset;

		spriteOffset = sprite * (16 * 16);

		for (int y = 15; y >= 0; y--)
		{
			s = &(srcBuffer[spriteOffset + (16 * y)]);
			f = &(flipBuffer[spriteOffset + (16 * (15 - y))]);
			memcpy(f,s,16);
		}
	}
}

/*-----------------------------------------------------------------------------
   Name: FlipSpritesHorizVert()
   Desc: Take decoded sprites and horizontally/vertically flip them
  -----------------------------------------------------------------------------*/
void Interpreter::FlipSpritesHorizVert(UCHAR *srcBuffer, UCHAR *flipBuffer)
{
	for (int sprite = 0; sprite < 64; sprite++)
	{
		int spriteOffset;

		spriteOffset = sprite * (16 * 16);

		for (int i = 0; i < (16 * 16); i++)
		{
			flipBuffer[spriteOffset + i] = srcBuffer[spriteOffset + ((16 * 16) - 1) - i];
		}
	}
}

/*-----------------------------------------------------------------------------
   Name: InitSprites()
   Desc: Initialize the interpreted sprite list
  -----------------------------------------------------------------------------*/
void Interpreter::InitSprites(void)
{
	for (int sprite = 0; sprite < NUM_SPRITES; sprite++)
	{
		spriteList[sprite].x = 0;
		spriteList[sprite].y = 0;
		spriteList[sprite].color = 0;
		spriteList[sprite].graphicsPtr = NULL;
	}
}

/*-----------------------------------------------------------------------------
   Name: Interpret()
   Desc: Interpret emulated memory into usable format
  -----------------------------------------------------------------------------*/
void Interpreter::Interpret(void)
{
	InterpretSprites();

	InterpretSound();
}

/*-----------------------------------------------------------------------------
   Name: InterpretSprites()
   Desc: Read emulated memory and refresh the interpreted sprite list
  -----------------------------------------------------------------------------*/
void Interpreter::InterpretSprites(void)
{
	int pic, flip;
	SPRITE_STRUCT *curSprite;

	curSprite = spriteList;

	// Flip sprite order for correct drawing priority
	for (int sprite = NUM_SPRITES - 1; sprite >= 0; sprite--)
	{
		curSprite->x = 255 - emuMemory[0x5060 + sprite*2];
		curSprite->y = 255 - emuMemory[0x5061 + sprite*2];
		curSprite->color = emuMemory[0x4ff1 + sprite*2] & 0x1F;
		pic   = emuMemory[0x4ff0 + sprite*2] >> 2;
		flip  = emuMemory[0x4ff0 + sprite*2] & 0x03;

		switch (flip)
		{
			/* No flip */
			case 0:
				curSprite->graphicsPtr = &(spritePtr[pic * (16*16)]);
				break;
			/* Vert flip */
			case 1:
				curSprite->graphicsPtr = &(spriteVFlipPtr[pic * (16*16)]);
				break;
			/* Horiz flip */
			case 2:
				curSprite->graphicsPtr = &(spriteHFlipPtr[pic * (16*16)]);
				break;
			/* Horiz/Vert flip */
			case 3:
				curSprite->graphicsPtr = &(spriteHVFlipPtr[pic * (16*16)]);
				break;
		}

		curSprite++;
	}
}

/*-----------------------------------------------------------------------------
   Name: InterpretSound()
   Desc: Read emulated memory and refresh sound parameters
  -----------------------------------------------------------------------------*/
void Interpreter::InterpretSound(void)
{
	if (emuMemory[0x5001] == 0)
		soundEnable = false;
	else
		soundEnable = true;

	for (int i = 0; i < NUM_SOUND_VOICES; i++ )
	{
		// Only mem locations 5045, 504A, 504F, and 5050-505F are used.

		wave[i]   = emuMemory[0x5045 + i*5] & 0x07;
		volume[i] = emuMemory[0x5055 + i*5] & 0x0F;

		frequency[i]  = (emuMemory[0x5054 + i*5] & 0x0F) << 16;
		frequency[i] |= (emuMemory[0x5053 + i*5] & 0x0F) << 12;
		frequency[i] |= (emuMemory[0x5052 + i*5] & 0x0F) << 8;
		frequency[i] |= (emuMemory[0x5051 + i*5] & 0x0F) << 4;

		// Special case - first voice gets an extra nibble of accuracy
		if (i == 0)
			frequency[i] |= (emuMemory[0x5050] & 0x0F);
	}
}
